Require Import Map.
Require Import TM.
Require Import Assertions.
Require Import SAGL.

Inductive CSL_Spec : Type :=
  | csl_spec (v : Assertion)
  | csl_nospec.

Definition CSL_CdHpSpec := Address -> CSL_Spec.

Definition LockSpec := Lock -> MemPred.

Inductive CSL_WFInstr (Psi : CSL_CdHpSpec) (Gamma : LockSpec)
  : CSL_Spec -> Instr -> CSL_Spec -> Prop :=
  | CSL_WFInstr_addu : 
    forall (v v' : Assertion) (rd rs rt : Reg),
      (forall M R k L M' R', 
	 Next (addu rd rs rt) (M, R) (M', R')
           -> v (M, (R, k), L) 
	      -> v' (M', (R', k), L))
      -> CSL_WFInstr Psi Gamma (csl_spec v) (addu rd rs rt) (csl_spec v')
  | CSL_WFInstr_cons :
    forall (v v' : Assertion) (rd : Reg),
      (forall M R k L,
	v (M, (R, k), L) ->
	forall l, not_in_dom l M -> not_in_dom (S l) M ->
	  forall w w' X,
	    X = (ext_M (ext_M M l w) (S l) w', (ext_R R rd l, k), L) -> v' X
      )
      -> CSL_WFInstr Psi Gamma (csl_spec v) (cons rd) (csl_spec v')
  | CSL_WFInstr_free :
    forall (v v' : Assertion) (rs : Reg),
      (forall M R k L,
	v (M, (R, k), L)
	  -> exists M', exists R', 
	        Next (free rs) (M, R) (M', R') /\ v' (M', (R', k), L)
      )
      -> CSL_WFInstr Psi Gamma (csl_spec v) (free rs) (csl_spec v')
  | CSL_WFInstr_sw :
    forall (v v' : Assertion) (rs rt: Reg) (w : Word),
      (forall M R k L,
	v (M, (R, k), L)
	  -> exists M', exists R', 
	        Next (sw rt w rs) (M, R) (M', R') /\ v' (M', (R', k), L)
      )
      -> CSL_WFInstr Psi Gamma (csl_spec v) (sw rt w rs) (csl_spec v')
  | CSL_WFInstr_lock :
    forall (v v' : Assertion) (l : Lock) (m : MemPred),
      Gamma l = m -> 
      (forall M R k L,
	(star_am v m) (M, (R, k), L) -> v' (M, (R, k), ins_L L l k)
      )
      -> CSL_WFInstr Psi Gamma (csl_spec v) (lock l) (csl_spec v')
  | CSL_WFInstr_unlock :
    forall (v v' : Assertion) (l : Lock) (m : MemPred),
      Gamma l = m -> 
      (forall M R k L,
	v (M, (R, k), L) -> (maps_to L l k) /\ (star_am v' m) (M, (R, k), del_L L l)
      )
      -> CSL_WFInstr Psi Gamma (csl_spec v) (unlock l) (csl_spec v')
  | CSL_WFInstr_bgtz :
    forall (v v' v'' : Assertion) (rs : Reg) (f : Address),
      Psi f = csl_spec v'' ->
      (forall  M R k L,
	v (M, (R, k), L)
	  -> (R rs > 0 -> v'' (M, (R, k), L)) /\
	     (R rs <= 0 -> v' (M, (R, k), L))
      )
      -> CSL_WFInstr Psi Gamma (csl_spec v) (bgtz rs f) (csl_spec v').

Inductive CSL_WFIseq (Psi : CSL_CdHpSpec) (Gamma: LockSpec)
  : CSL_Spec -> InstrSeq -> Prop :=
  | CSL_WFIseq_seq : 
    forall (v v' : Assertion) (c : Instr) (I : InstrSeq),
      CSL_WFInstr Psi Gamma (csl_spec v) c (csl_spec v')
        -> CSL_WFIseq Psi Gamma (csl_spec v') I
	-> CSL_WFIseq Psi Gamma (csl_spec v) (seq c I)
  | CSL_WFIseq_jr : 
    forall (v v' : Assertion) (rs : Reg),
      (forall M R k L,
	v (M, (R, k), L) -> (Psi (R rs) = (csl_spec v')) /\ v' (M, (R, k), L))
      -> CSL_WFIseq Psi Gamma (csl_spec v) (jr rs).

Definition CSL_WFcdhp (Psi : CSL_CdHpSpec) (Gamma : LockSpec) (C : CodeHeap) (Psi' : CSL_CdHpSpec) : Prop :=
      forall (f : Address) (v : Assertion),
	Psi' f = csl_spec v -> exists I, maps_to C f I /\ CSL_WFIseq Psi Gamma (csl_spec v) I.

Definition CSL_WFthrd (Psi : CSL_CdHpSpec) (Gamma : LockSpec) (v : Assertion) 
                       (T : Thread) : Prop :=
  match T with
    | (C, R, IS, k) => CSL_WFcdhp Psi Gamma C Psi /\ CSL_WFIseq Psi Gamma (csl_spec v) IS
  end.

Definition forallstar_m (L : LockMap) (Gamma : LockSpec) : MemPred :=
  star_m (fun M => (not_in_dom l1 L -> Gamma l1 M) /\ (in_dom l1 L -> emp_m M))
         (fun M => (not_in_dom l2 L -> Gamma l2 M) /\ (in_dom l2 L -> emp_m M)).

(* the a_Gamma in our paper *)
Definition itp_Gamma (Gamma : LockSpec): Assertion :=
  fun (X : XState) =>
    match X with
      (M, (R, k), L) => forallstar_m L Gamma M
    end.

Definition deno_Psi (Psy : CSL_CdHpSpec) (Gamma : LockSpec) : SAGL_CdHpSpec :=
  fun f => match Psy f with
           | csl_spec v => sagl_spec (itp_Gamma Gamma) v
           | csl_nospec => sagl_nospec
           end.

Definition assume_Gamma (Gamma : LockSpec) : ASSUME :=
  fun X X' => match X,X' with
              | (M,(R,k),L),(M',(R',k'),L') => R = R' /\ k = k' /\ (itp_Gamma Gamma (M,(R,k),L) -> itp_Gamma Gamma (M',(R',k'),L'))
              end.

Definition grt_Gamma (Gamma : LockSpec) : GRT :=
  fun X X' => match X,X' with
              | (M,(R,k),L),(M',(R',k'),L') => k = k' /\ itp_Gamma Gamma (M,(R,k),L) /\ itp_Gamma Gamma (M',(R',k'),L')
              end.

Definition CSL_WFlockspec (Gamma : LockSpec) : Prop :=
  forall (l : Lock), prec_m (Gamma l).

Definition CSL_WFprog (Psi Psi' : CSL_CdHpSpec) (Gamma : LockSpec)
                       (v v': Assertion) (P : Program) : Prop :=
  (match P with
    | (M, (T_1, T_2), L) =>
      match T_1 with (C_1, R_1, I_1, tid_1) =>
	match T_2 with (C_2, R_2, I_2, tid_2) =>
          tid_1 <> tid_2 /\
	  exists MS, exists M', exists M_1, exists M_2,
	    disjUnion M_1 M_2 M' /\ disjUnion M' MS M /\
	    v (M_1, (R_1, tid_1), filter_L L tid_1) /\
	    v' (M_2, (R_2, tid_2), filter_L L tid_2) /\
	    (CSL_WFthrd Psi Gamma v T_1) /\
	    (CSL_WFthrd Psi' Gamma v' T_2) /\ 
	    (forallstar_m L Gamma MS)
	end
      end
   end) /\ CSL_WFlockspec Gamma.

Definition star_Psi (Psi : CSL_CdHpSpec) (m : MemPred) : CSL_CdHpSpec :=
  fun f => match Psi f with
           | csl_spec v => csl_spec (star_am v m)
           | csl_nospec => csl_nospec
           end.

